﻿using System;
using System.Configuration;
using System.Configuration.Provider;
using System.Data;
using System.Diagnostics;
using System.Security.Cryptography;
using System.Text;
using System.Web.Configuration;
using System.Web.Hosting;
using System.Web.Security;
using VistaDB;
using VistaDB.Diagnostic;
using VistaDB.Provider;

namespace VistaDBProviders.MembershipProvider
{
    public class VistaDBMembership : System.Web.Security.MembershipProvider
    {
        private string connectionString;
        private string eventLog = "Application";
        private string eventSource = "VistaDBMembershipProvider";
        private string exceptionMessage = "An exception occurred. Please check the Event Log.";
        private MachineKeySection machineKey;
        private int newPasswordLength = 8;
        private string pApplicationName;
        private bool pEnablePasswordReset;
        private bool pEnablePasswordRetrieval;
        private int pMaxInvalidPasswordAttempts;
        private int pMinRequiredNonAlphanumericCharacters;
        private int pMinRequiredPasswordLength;
        private int pPasswordAttemptWindow;
        private MembershipPasswordFormat pPasswordFormat;
        private string pPasswordStrengthRegularExpression;
        private bool pRequiresQuestionAndAnswer;
        private bool pRequiresUniqueEmail;
        private bool pWriteExceptionsToEventLog;
		
        public override string ApplicationName
        {
            get { return pApplicationName; }
            set { pApplicationName = value; }
        }

        public override bool EnablePasswordReset
        {
            get { return pEnablePasswordReset; }
        }

        public override bool EnablePasswordRetrieval
        {
            get { return pEnablePasswordRetrieval; }
        }

        public override int MaxInvalidPasswordAttempts
        {
            get { return pMaxInvalidPasswordAttempts; }
        }

        public override int MinRequiredNonAlphanumericCharacters
        {
            get { return pMinRequiredNonAlphanumericCharacters; }
        }

        public override int MinRequiredPasswordLength
        {
            get { return pMinRequiredPasswordLength; }
        }

        public override int PasswordAttemptWindow
        {
            get { return pPasswordAttemptWindow; }
        }

        public override MembershipPasswordFormat PasswordFormat
        {
            get { return pPasswordFormat; }
        }

        public override string PasswordStrengthRegularExpression
        {
            get { return pPasswordStrengthRegularExpression; }
        }

        public override bool RequiresQuestionAndAnswer
        {
            get { return pRequiresQuestionAndAnswer; }
        }

        public override bool RequiresUniqueEmail
        {
            get { return pRequiresUniqueEmail; }
        }

        public bool WriteExceptionsToEventLog
        {
            get { return pWriteExceptionsToEventLog; }
            set { pWriteExceptionsToEventLog = value; }
        }

        public override bool ChangePassword(string username, string oldPassword, string newPassword)
        {
            if (!ValidateUser(username, oldPassword))
                return false;

            ValidatePasswordEventArgs args = new ValidatePasswordEventArgs(username, newPassword, true);
            OnValidatingPassword(args);

            if (args.Cancel)
            {
                if (args.FailureInformation != null)
                    throw args.FailureInformation;
                else
                    throw new ProviderException("Change password canceled due to New password validation failure.");
            }

            VistaDBConnection conn = new VistaDBConnection(connectionString);
            VistaDBCommand cmd = new VistaDBCommand("UPDATE Users " +
                                " SET Password = @Password, LastPasswordChangedDate = @LastPasswordChangedDate " +
                                " WHERE Username = @Username AND ApplicationName = @ApplicationName", conn);

            cmd.Parameters.Add("@Password", VistaDBType.VarChar, 255).Value = EncodePassword(newPassword);
            cmd.Parameters.Add("@LastPasswordChangedDate", VistaDBType.DateTime).Value = DateTime.Now;
            cmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = username;
            cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

            int rowsAffected = 0;

            try
            {

                conn.Open();
                rowsAffected = cmd.ExecuteNonQuery();
            }
            catch (VistaDBException e)
            {
                if (WriteExceptionsToEventLog)
                {
                    WriteToEventLog(e, "ChangePassword");
                    throw new ProviderException(exceptionMessage);
                }
                else
                    throw e;
            }
            finally
            {
                conn.Close();
            }
            if (rowsAffected > 0)
                return true;

            return false;
        }

        public override bool ChangePasswordQuestionAndAnswer(string username, string password, string newPasswordQuestion, string newPasswordAnswer)
        {
            if (!ValidateUser(username, password))
                return false;

            VistaDBConnection conn = new VistaDBConnection(connectionString);
            VistaDBCommand cmd = new VistaDBCommand("UPDATE Users " +
                    " SET PasswordQuestion = @Question, PasswordAnswer = @Answer" +
                    " WHERE Username = @Username AND ApplicationName = @ApplicationName", conn);

            cmd.Parameters.Add("@Question", VistaDBType.VarChar, 255).Value = newPasswordQuestion;
            cmd.Parameters.Add("@Answer", VistaDBType.VarChar, 255).Value = EncodePassword(newPasswordAnswer);
            cmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = username;
            cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

            int rowsAffected = 0;

            try
            {
                conn.Open();

                rowsAffected = cmd.ExecuteNonQuery();
            }
            catch (VistaDBException e)
            {
                if (WriteExceptionsToEventLog)
                {
                    WriteToEventLog(e, "ChangePasswordQuestionAndAnswer");
                    throw new ProviderException(exceptionMessage);
                }
                else
                    throw e;
            }
            finally
            {
                conn.Close();
            }

            if (rowsAffected > 0)
                return true;

            return false;
        }

        public override MembershipUser CreateUser(string username, string password, string email, string passwordQuestion, string passwordAnswer, bool isApproved, object providerUserKey, out MembershipCreateStatus status)
        {
            ValidatePasswordEventArgs Args =
              new ValidatePasswordEventArgs(username, password, true);

            OnValidatingPassword(Args);

            if (Args.Cancel)
            {
                status = MembershipCreateStatus.InvalidPassword;
                return null;
            }


            if (RequiresUniqueEmail && GetUserNameByEmail(email) != "")
            {
                status = MembershipCreateStatus.DuplicateEmail;
                return null;
            }

            MembershipUser u = GetUser(username, false);

            if (u == null)
            {
                DateTime createDate = DateTime.Now;

                if (providerUserKey == null)
                    providerUserKey = Guid.NewGuid();
                else
                {
                    if (providerUserKey is Guid)
                    {
                        status = MembershipCreateStatus.InvalidProviderUserKey;
                        return null;
                    }
                }

                VistaDBConnection conn = new VistaDBConnection(connectionString);
                VistaDBCommand cmd = new VistaDBCommand("INSERT INTO Users " +
                       " (PKID, Username, Password, Email, PasswordQuestion, " +
                       " PasswordAnswer, IsApproved," +
                       " Comment, CreationDate, LastPasswordChangedDate, LastActivityDate," +
                       " ApplicationName, IsLockedOut, LastLockedOutDate," +
                       " FailedPasswordAttemptCount, FailedPasswordAttemptWindowStart, " +
                       " FailedPasswordAnswerAttemptCount, FailedPasswordAnswerAttemptWindowStart)" +
                       " Values(@PKID, @Username, @Password, @Email, @PasswordQuestion, @PasswordAnswer, @IsApproved, @Comment" +
                       ", @CreationDate, @LastPasswordChangedDate, @LastActivityDate, @ApplicationName, @IsLockedOut, @LastLockedOutDate, @FailedPasswordAttemptCount, @FailedPasswordAttemptWindowStart, @FailedPasswordAnswerAttemptCount, @FailedPasswordAnswerAttemptWindowStart)", conn);

                cmd.Parameters.Add("@PKID", VistaDBType.UniqueIdentifier).Value = providerUserKey;
                cmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = username;
                cmd.Parameters.Add("@Password", VistaDBType.VarChar, 255).Value = EncodePassword(password);
                cmd.Parameters.Add("@Email", VistaDBType.VarChar, 128).Value = email;
                cmd.Parameters.Add("@PasswordQuestion", VistaDBType.VarChar, 255).Value = passwordQuestion;
                cmd.Parameters.Add("@PasswordAnswer", VistaDBType.VarChar, 255).Value = EncodePassword(passwordAnswer);
                cmd.Parameters.Add("@IsApproved", VistaDBType.Bit).Value = isApproved;
                cmd.Parameters.Add("@Comment", VistaDBType.VarChar, 255).Value = "";
                cmd.Parameters.Add("@CreationDate", VistaDBType.DateTime).Value = createDate;
                cmd.Parameters.Add("@LastPasswordChangedDate", VistaDBType.DateTime).Value = createDate;
                cmd.Parameters.Add("@LastActivityDate", VistaDBType.DateTime).Value = createDate;
                cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;
                cmd.Parameters.Add("@IsLockedOut", VistaDBType.Bit).Value = false;
                cmd.Parameters.Add("@LastLockedOutDate", VistaDBType.DateTime).Value = createDate;
                cmd.Parameters.Add("@FailedPasswordAttemptCount", VistaDBType.Int).Value = 0;
                cmd.Parameters.Add("@FailedPasswordAttemptWindowStart", VistaDBType.DateTime).Value = createDate;
                cmd.Parameters.Add("@FailedPasswordAnswerAttemptCount", VistaDBType.Int).Value = 0;
                cmd.Parameters.Add("@FailedPasswordAnswerAttemptWindowStart", VistaDBType.DateTime).Value = createDate;


                try
                {
                    conn.Open();

                    int recAdded = cmd.ExecuteNonQuery();

                    if (recAdded > 0)
                        status = MembershipCreateStatus.Success;
                    else
                        status = MembershipCreateStatus.UserRejected;
                }
                catch (VistaDBException e)
                {
                    if (WriteExceptionsToEventLog)
                        WriteToEventLog(e, "CreateUser");

                    status = MembershipCreateStatus.ProviderError;
                }
                finally
                {
                    conn.Close();
                }
                return GetUser(username, false);
            }
            else
                status = MembershipCreateStatus.DuplicateUserName;

            return null;
        }

        public override bool DeleteUser(string username, bool deleteAllRelatedData)
        {
            VistaDBConnection conn = new VistaDBConnection(connectionString);
            VistaDBCommand cmd = new VistaDBCommand("DELETE FROM Users " +
                    " WHERE Username = @Username AND Applicationname = @ApplicationName", conn);

            cmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = username;
            cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

            int rowsAffected = 0;

            try
            {
                conn.Open();

                rowsAffected = cmd.ExecuteNonQuery();
                if (deleteAllRelatedData)
                {
                    VistaDBCommand cmd2 = new VistaDBCommand("DELETE FROM Usersinroles " +
                    " WHERE Username = @Username AND Applicationname = @ApplicationName", conn);

                    cmd2.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = username;
                    cmd2.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

                    cmd2.ExecuteNonQuery();
                }
            }
            catch (VistaDBException e)
            {
                if (WriteExceptionsToEventLog)
                {
                    WriteToEventLog(e, "DeleteUser");
                    throw new ProviderException(exceptionMessage);
                }
                else
                    throw e;
            }
            finally
            {
                conn.Close();
            }

            if (rowsAffected > 0)
                return true;

            return false;
        }

        public override MembershipUserCollection FindUsersByEmail(string emailToMatch, int pageIndex, int pageSize, out int totalRecords)
        {
            VistaDBConnection conn = new VistaDBConnection(connectionString);
            VistaDBCommand cmd = new VistaDBCommand("SELECT Count(*) FROM Users  " +
                                          "WHERE Email LIKE @EmailSearch AND ApplicationName = @ApplicationName", conn);

        cmd.Parameters.Add("@EmailSearch", VistaDBType.VarChar, 255).Value = emailToMatch;
        cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = ApplicationName;

            MembershipUserCollection users = new MembershipUserCollection();

            VistaDBDataReader reader = null;

            try
            {
                conn.Open();
                totalRecords = (int)(cmd.ExecuteScalar());

                if (totalRecords <= 0)
                    return users;

                cmd.CommandText = "SELECT PKID, Username, Email, PasswordQuestion," +
                     " Comment, IsApproved, IsLockedOut, CreationDate, LastLoginDate," +
                     " LastActivityDate, LastPasswordChangedDate, LastLockedOutDate " +
                     " FROM Users  " +
                     " WHERE Email LIKE @EmailSearch AND ApplicationName = @ApplicationName " +
                     " ORDER BY Username Asc";

                reader = cmd.ExecuteReader();

                int counter = 0;
                int startIndex = pageSize * pageIndex;
                int endIndex = startIndex + pageSize - 1;

                while (reader.Read())
                {
                    if (counter >= startIndex)
                    {
                        MembershipUser u = GetUserFromReader(reader);
                        users.Add(u);
                    }
                    if (counter >= endIndex)
                        cmd.Cancel();

                    counter += 1;
                }
            }
            catch (VistaDBException e)
            {
                if (WriteExceptionsToEventLog)
                {
                    WriteToEventLog(e, "FindUsersByName");
                    throw new ProviderException(exceptionMessage);
                }
                else
                    throw e;
            }
            finally
            {
                if (reader != null)
                    reader.Close();
                conn.Close();
            }

            return users;
        }

        public override MembershipUserCollection FindUsersByName(string usernameToMatch, int pageIndex, int pageSize, out int totalRecords)
        {
            VistaDBConnection conn = new VistaDBConnection(connectionString);
            VistaDBCommand cmd = new VistaDBCommand("SELECT Count(*) FROM Users  " +
                      "WHERE Username LIKE @UsernameSearch AND ApplicationName = @ApplicationName", conn);

            cmd.Parameters.Add("@UsernameSearch", VistaDBType.VarChar, 255).Value = usernameToMatch;
            cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

            MembershipUserCollection users = new MembershipUserCollection();

            VistaDBDataReader reader = null;

            try
            {
                conn.Open();
                totalRecords = (int)(cmd.ExecuteScalar());

                if (totalRecords <= 0)
                    return users;

                cmd.CommandText = "SELECT PKID, Username, Email, PasswordQuestion," +
                  " Comment, IsApproved, IsLockedOut, CreationDate, LastLoginDate," +
                  " LastActivityDate, LastPasswordChangedDate, LastLockedOutDate " +
                  " FROM Users  " +
                  " WHERE Username LIKE @UsernameSearch AND ApplicationName = @ApplicationName " +
                  " ORDER BY Username Asc";

                reader = cmd.ExecuteReader();

                int counter = 0;
                int startIndex = pageSize * pageIndex;
                int endIndex = startIndex + pageSize - 1;

                while (reader.Read())
                {
                    if (counter >= startIndex)
                    {
                        MembershipUser u = GetUserFromReader(reader);
                        users.Add(u);
                    }
                    if (counter >= endIndex)
                        cmd.Cancel();

                    counter += 1;
                }
            }
            catch (VistaDBException e)
            {
                if (WriteExceptionsToEventLog)
                {
                    WriteToEventLog(e, "FindUsersByName");
                    throw new ProviderException(exceptionMessage);
                }
                else
                    throw e;
            }
            finally
            {
                if (reader != null)
                    reader.Close();
                conn.Close();
            }

            return users;
        }

        public override MembershipUserCollection GetAllUsers(int pageIndex, int pageSize, out int totalRecords)
        {
            VistaDBConnection conn = new VistaDBConnection(connectionString);
            VistaDBCommand cmd = new VistaDBCommand("SELECT Count(*) FROM Users  " +
                                              "WHERE ApplicationName = @ApplicationName", conn);
            cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = ApplicationName;

            MembershipUserCollection users = new MembershipUserCollection();

            VistaDBDataReader reader = null;
            totalRecords = 0;

            try
            {
                conn.Open();
                totalRecords = (int)(cmd.ExecuteScalar());

                if (totalRecords <= 0)
                    return users;

                cmd.CommandText = "SELECT PKID, Username, Email, PasswordQuestion," +
                         " Comment, IsApproved, IsLockedOut, CreationDate, LastLoginDate," +
                         " LastActivityDate, LastPasswordChangedDate, LastLockedOutDate " +
                         " FROM Users  " +
                         " WHERE ApplicationName = @ApplicationName " +
                         " ORDER BY Username Asc";

                reader = cmd.ExecuteReader();

                int counter = 0;
                int startIndex = pageSize * pageIndex;
                int endIndex = startIndex + pageSize - 1;

                while (reader.Read())
                {
                    if (counter >= startIndex)
                    {
                        MembershipUser u = GetUserFromReader(reader);
                        users.Add(u);
                    }

                    if (counter >= endIndex)
                        cmd.Cancel();

                    counter += 1;
                }
            }
            catch (VistaDBException e)
            {
                if (WriteExceptionsToEventLog)
                {
                    WriteToEventLog(e, "GetAllUsers");
                    throw new ProviderException(exceptionMessage);
                }
                else
                    throw e;
            }
            finally
            {
                if (reader != null)
                    reader.Close();
                conn.Close();
            }

            return users;
        }

        public override int GetNumberOfUsersOnline()
        {
            TimeSpan onlineSpan = new TimeSpan(0, Membership.UserIsOnlineTimeWindow, 0);
            DateTime compareTime = DateTime.Now.Subtract(onlineSpan);

            VistaDBConnection conn = new VistaDBConnection(connectionString);
            VistaDBCommand cmd = new VistaDBCommand("SELECT Count(*) FROM Users " +
                    " WHERE LastActivityDate > @CompareDate AND ApplicationName = @ApplicationName", conn);

            cmd.Parameters.Add("@CompareDate", VistaDBType.DateTime).Value = compareTime;
            cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

            int numOnline = 0;

            try
            {
                conn.Open();
                numOnline = (int)(cmd.ExecuteScalar());
            }
            catch (VistaDBException e)
            {
                if (WriteExceptionsToEventLog)
                {
                    WriteToEventLog(e, "GetNumberOfUsersOnline");
                    throw new ProviderException(exceptionMessage);
                }
                else
                    throw e;
            }
            finally
            {
                conn.Close();
            }

            return numOnline;
        }

        public override string GetPassword(string username, string answer)
        {
            if (!EnablePasswordRetrieval)
                throw new ProviderException("Password Retrieval Not Enabled.");

            if (PasswordFormat == MembershipPasswordFormat.Hashed)
                throw new ProviderException("Cannot retrieve Hashed passwords.");

            VistaDBConnection conn = new VistaDBConnection(connectionString);
            VistaDBCommand cmd = new VistaDBCommand("SELECT Password, PasswordAnswer, IsLockedOut FROM Users " +
                  " WHERE Username = @Username AND ApplicationName = @ApplicationName", conn);

            cmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = username;
            cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

            string password = "";
            string passwordAnswer = "";
            VistaDBDataReader reader = null;

            try
            {
                conn.Open();
                reader = cmd.ExecuteReader(CommandBehavior.SingleRow);

                if (reader.HasRows)
                {
                    reader.Read();

                    if (reader.GetBoolean(2))
                        throw new MembershipPasswordException("The supplied user is locked out.");

                    password = reader.GetString(0);
                    passwordAnswer = reader.GetString(1);
                }
                else
                    throw new MembershipPasswordException("The supplied user name is not found.");
            }
            catch (VistaDBException e)
            {
                if (WriteExceptionsToEventLog)
                {
                    WriteToEventLog(e, "GetPassword");
                    throw new ProviderException(exceptionMessage);
                }
                else
                    throw e;
            }
            finally
            {
                if (reader != null)
                    reader.Close();
                conn.Close();
            }

            if (RequiresQuestionAndAnswer && !CheckPassword(answer, passwordAnswer))
            {
                UpdateFailureCount(username, "passwordAnswer");
                throw new MembershipPasswordException("Incorrect password answer.");
            }

            if (PasswordFormat == MembershipPasswordFormat.Encrypted)
                password = UnEncodePassword(password);

            return password;
        }

        public override MembershipUser GetUser(string username, bool userIsOnline)
        {
            VistaDBConnection conn = new VistaDBConnection(connectionString);
            VistaDBCommand cmd = new VistaDBCommand("SELECT PKID, Username, Email, PasswordQuestion," +
                  " Comment, IsApproved, IsLockedOut, CreationDate, LastLoginDate," +
                  " LastActivityDate, LastPasswordChangedDate, LastLockedOutDate" +
                  " FROM Users  WHERE Username = @Username AND ApplicationName = @ApplicationName", conn);

            cmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = username;
            cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

            MembershipUser u = null;
            VistaDBDataReader reader = null;

            try
            {
                conn.Open();
                reader = cmd.ExecuteReader();

                if (reader.HasRows)
                {
                    reader.Read();
                    u = GetUserFromReader(reader);

                    if (userIsOnline)
                    {
                        VistaDBCommand updateCmd = new VistaDBCommand("UPDATE Users  " +
                                     "SET LastActivityDate = @LastActivityDate " +
                                     "WHERE Username = @Username AND Applicationname = @ApplicationName", conn);

                        updateCmd.Parameters.Add("@LastActivityDate", VistaDBType.DateTime).Value = DateTime.Now;
                        updateCmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = username;
                        updateCmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

                        updateCmd.ExecuteNonQuery();
                    }
                }
            }
            catch (VistaDBException e)
            {
                if (WriteExceptionsToEventLog)
                {
                    WriteToEventLog(e, "GetUser(String, Boolean)");
                    throw new ProviderException(exceptionMessage);
                }
                else
                    throw e;

            }
            finally
            {
                if (reader != null)
                    reader.Close();
                conn.Close();
            }

            return u;
        }

        public override MembershipUser GetUser(object providerUserKey, bool userIsOnline)
        {
            VistaDBConnection conn = new VistaDBConnection(connectionString);
            VistaDBCommand cmd = new VistaDBCommand("SELECT PKID, Username, Email, PasswordQuestion," +
                  " Comment, IsApproved, IsLockedOut, CreationDate, LastLoginDate," +
                  " LastActivityDate, LastPasswordChangedDate, LastLockedOutDate" +
                  " FROM Users  WHERE PKID = @PKID", conn);

            cmd.Parameters.Add("@PKID", VistaDBType.UniqueIdentifier).Value = providerUserKey;

            MembershipUser u = null;
            VistaDBDataReader reader = null;

            try
            {
                conn.Open();
                reader = cmd.ExecuteReader();

                if (reader.HasRows)
                {
                    reader.Read();
                    u = GetUserFromReader(reader);

                    if (userIsOnline)
                    {
                        VistaDBCommand updateCmd = new VistaDBCommand("UPDATE Users  " +
                                  "SET LastActivityDate = @LastActivityDate " +
                                  "WHERE PKID = @PKID", conn);

                        updateCmd.Parameters.Add("@LastActivityDate", VistaDBType.DateTime).Value = DateTime.Now;
                        updateCmd.Parameters.Add("@PKID", VistaDBType.UniqueIdentifier).Value = providerUserKey;

                        updateCmd.ExecuteNonQuery();
                    }
                }
            }
            catch (VistaDBException e)
            {
                if (WriteExceptionsToEventLog)
                {
                    WriteToEventLog(e, "GetUser(Object, Boolean)");
                    throw new ProviderException(exceptionMessage);
                }
                else
                    throw e;
            }
            finally
            {
                if (reader != null)
                    reader.Close();
                conn.Close();
            }

            return u;
        }

        public override string GetUserNameByEmail(string email)
        {
            VistaDBConnection conn = new VistaDBConnection(connectionString);
            VistaDBCommand cmd = new VistaDBCommand("SELECT Username" +
                  " FROM Users  WHERE Email = @Email AND ApplicationName = @ApplicationName", conn);

            cmd.Parameters.Add("@Email", VistaDBType.VarChar, 128).Value = email;
            cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

            string username = "";

            try
            {
                conn.Open();
                username = (string)cmd.ExecuteScalar();
            }
            catch (VistaDBException e)
            {
                if (WriteExceptionsToEventLog)
                {
                    WriteToEventLog(e, "GetUserNameByEmail");
                    throw new ProviderException(exceptionMessage);
                }
                else
                    throw e;
            }
            finally
            {
                conn.Close();
            }

            if (username == null)
                username = "";

            return username;
        }

        public override void Initialize(string name, System.Collections.Specialized.NameValueCollection config)
        {
            if (config == null)
                throw new ArgumentNullException("Config section null");

            if (name == null || name.Length == 0)
                name = "VistaDBMembership";

            if (string.IsNullOrEmpty(config["description"]))
            {
                config.Remove("description");
                config.Add("description", "VistaDB Membership provider");
            }

            base.Initialize(name, config);

            pApplicationName = GetConfigValue(config["applicationName"], HostingEnvironment.ApplicationVirtualPath);
            pMaxInvalidPasswordAttempts = Convert.ToInt32(GetConfigValue(config["maxInvalidPasswordAttempts"], "5"));
            pPasswordAttemptWindow = Convert.ToInt32(GetConfigValue(config["passwordAttemptWindow"], "10"));
            pMinRequiredNonAlphanumericCharacters = Convert.ToInt32(GetConfigValue(config["minRequiredAlphaNumericCharacters"], "1"));
            pMinRequiredPasswordLength = Convert.ToInt32(GetConfigValue(config["minRequiredPasswordLength"], "7"));
            pPasswordStrengthRegularExpression = Convert.ToString(GetConfigValue(config["passwordStrengthRegularExpression"], ""));
            pEnablePasswordReset = Convert.ToBoolean(GetConfigValue(config["enablePasswordReset"], "True"));
            pEnablePasswordRetrieval = Convert.ToBoolean(GetConfigValue(config["enablePasswordRetrieval"], "True"));
            pRequiresQuestionAndAnswer = Convert.ToBoolean(GetConfigValue(config["requiresQuestionAndAnswer"], "False"));
            pRequiresUniqueEmail = Convert.ToBoolean(GetConfigValue(config["requiresUniqueEmail"], "True"));
            pWriteExceptionsToEventLog = Convert.ToBoolean(GetConfigValue(config["writeExceptionsToEventLog"], "True"));

            string temp_format = config["passwordFormat"];
            if (temp_format == null)
                temp_format = "Hashed";

            switch (temp_format)
            {
                case "Hashed":
                    pPasswordFormat = MembershipPasswordFormat.Hashed;
                    break;
                case "Encrypted":
                    pPasswordFormat = MembershipPasswordFormat.Encrypted;
                    break;
                case "Clear":
                    pPasswordFormat = MembershipPasswordFormat.Clear;
                    break;
                default:
                    throw new ProviderException("Password format not supported.");
                    break;
            }

            ConnectionStringSettings connectionStringSettings =
                ConfigurationManager.ConnectionStrings[config["connectionStringName"]];

            if (connectionStringSettings == null || connectionStringSettings.ConnectionString.Trim() == "")
                throw new ProviderException("Connection string cannot be blank.");

            connectionString = connectionStringSettings.ConnectionString;

            //Configuration cfg =
            //    WebConfigurationManager.OpenMachineConfiguration(HostingEnvironment.ApplicationVirtualPath);
            //machineKey = cfg.GetSection("system.web/machineKey") as MachineKeySection;

            //if (machineKey.ValidationKey.Contains("Autogenerate"))
            //{
            //    if (PasswordFormat != MembershipPasswordFormat.Clear)
            //        throw new ProviderException("Hashed or Encrypted passwords " +
            //                "are not supported with auto-generated keys.");
            //}
        }

        public override string ResetPassword(string username, string answer)
        {
            if (!EnablePasswordReset)
                throw new NotSupportedException("Password Reset is not enabled.");

            if (answer == null && RequiresQuestionAndAnswer)
            {
                UpdateFailureCount(username, "passwordAnswer");
                throw new ProviderException("Password answer required for password Reset.");
            }

            string newPassword =
              Membership.GeneratePassword(newPasswordLength, MinRequiredNonAlphanumericCharacters);


            ValidatePasswordEventArgs Args = new ValidatePasswordEventArgs(username, newPassword, true);
            OnValidatingPassword(Args);

            if (Args.Cancel)
            {
                if (Args.FailureInformation != null)
                    throw Args.FailureInformation;
                else
                    throw new MembershipPasswordException("Reset password canceled due to password validation failure.");
            }

            VistaDBConnection conn = new VistaDBConnection(connectionString);
            VistaDBCommand cmd = new VistaDBCommand("SELECT PasswordAnswer, IsLockedOut FROM Users " +
                  " WHERE Username = @Username AND ApplicationName = @ApplicationName", conn);

            cmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = username;
            cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

            int rowsAffected = 0;
            string passwordAnswer = "";
            VistaDBDataReader reader = null;

            try
            {
                conn.Open();
                reader = cmd.ExecuteReader(CommandBehavior.SingleRow);

                if (reader.HasRows)
                {
                    reader.Read();
                    if (reader.GetBoolean(1))
                        throw new MembershipPasswordException("The supplied user is locked out.");
                    passwordAnswer = reader.GetString(0);
                }
                else
                    throw new MembershipPasswordException("The supplied user name is not found.");

                if (RequiresQuestionAndAnswer && !CheckPassword(answer, passwordAnswer))
                {
                    UpdateFailureCount(username, "passwordAnswer");
                    throw new MembershipPasswordException("Incorrect password answer.");
                }

                VistaDBCommand updateCmd = new VistaDBCommand("UPDATE Users " +
                     " SET Password = @Password, LastPasswordChangedDate = @LastPasswordChangedDate" +
                     " WHERE Username = @Username AND ApplicationName = @ApplicationName AND IsLockedOut = False", conn);

                updateCmd.Parameters.Add("@Password", VistaDBType.VarChar, 255).Value = EncodePassword(newPassword);
                updateCmd.Parameters.Add("@LastPasswordChangedDate", VistaDBType.DateTime).Value = DateTime.Now;
                updateCmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = username;
                updateCmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

                rowsAffected = updateCmd.ExecuteNonQuery();
            }
            catch (VistaDBException e)
            {
                if (WriteExceptionsToEventLog)
                {
                    WriteToEventLog(e, "ResetPassword");
                    throw new ProviderException(exceptionMessage);
                }
                else
                    throw e;
            }
            finally
            {
                if (reader != null)
                    reader.Close();
                conn.Close();
            }

            if (rowsAffected > 0)
                return newPassword;
            else
                throw new MembershipPasswordException("User not found, or user is locked out. Password not Reset.");
        }

        public override bool UnlockUser(string userName)
        {
            VistaDBConnection conn = new VistaDBConnection(connectionString);
            VistaDBCommand cmd = new VistaDBCommand("UPDATE Users  " +
                                              " SET IsLockedOut = False, LastLockedOutDate = @LastLockedOutDate " +
                                              " WHERE Username = @Username AND ApplicationName = @ApplicationName", conn);

            cmd.Parameters.Add("@LastLockedOutDate", VistaDBType.DateTime).Value = DateTime.Now;
            cmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = userName;
            cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

            int rowsAffected = 0;

            try
            {
                conn.Open();
                rowsAffected = cmd.ExecuteNonQuery();
            }
            catch (VistaDBException e)
            {
                if (WriteExceptionsToEventLog)
                {
                    WriteToEventLog(e, "UnlockUser");
                    throw new ProviderException(exceptionMessage);
                }
                else
                    throw e;
            }
            finally
            {
                conn.Close();
            }

            if (rowsAffected > 0)
                return true;

            return false;
        }

        public override void UpdateUser(MembershipUser user)
        {
            VistaDBConnection conn = new VistaDBConnection(connectionString);
            VistaDBCommand cmd = new VistaDBCommand("UPDATE Users " +
                    " SET Email = @Email, Comment = @Comment," +
                    " IsApproved = @IsApproved" +
                    " WHERE Username = @Username AND ApplicationName = @ApplicationName", conn);

            cmd.Parameters.Add("@Email", VistaDBType.VarChar, 128).Value = user.Email;
            cmd.Parameters.Add("@Comment", VistaDBType.VarChar, 255).Value = user.Comment;
            cmd.Parameters.Add("@IsApproved", VistaDBType.Bit).Value = user.IsApproved;
            cmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = user.UserName;
            cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;


            try
            {
                conn.Open();
                cmd.ExecuteNonQuery();
            }
            catch (VistaDBException e)
            {
                if (WriteExceptionsToEventLog)
                {
                    WriteToEventLog(e, "UpdateUser");
                    throw new ProviderException(exceptionMessage);
                }
                else
                    throw e;
            }
            finally
            {
                conn.Close();
            }
        }

        public override bool ValidateUser(string username, string password)
        {
            bool isValid = false;

            VistaDBConnection conn = new VistaDBConnection(connectionString);
            VistaDBCommand cmd = new VistaDBCommand("SELECT Password, IsApproved FROM Users " +
                 " WHERE Username = @Username AND ApplicationName = @ApplicationName AND IsLockedOut = false", conn);

            cmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = username;
            cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

            VistaDBDataReader reader = null;
            bool isApproved = false;
            string pwd = "";

            try
            {
                conn.Open();
                reader = cmd.ExecuteReader(CommandBehavior.SingleRow);
                if (reader.HasRows)
                {
                    reader.Read();
                    pwd = reader["Password"].ToString();
                    isApproved = (bool)reader["IsApproved"];
                }
                else
                    return false;
                reader.Close();
                if (CheckPassword(password, pwd))
                {
                    if (isApproved)
                    {
                        isValid = true;
                        VistaDBCommand updateCmd = new VistaDBCommand("UPDATE Users  SET LastLoginDate = @LastLoginDate" +
                                                            " WHERE Username = @Username AND ApplicationName = @ApplicationName", conn);
                        updateCmd.Parameters.Add("@LastLoginDate", VistaDBType.DateTime).Value = DateTime.Now;
                        updateCmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = username;
                        updateCmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

                        updateCmd.ExecuteNonQuery();
                    }
                }
                else
                {
                    conn.Close();
                    UpdateFailureCount(username, "password");
                }
            }
            catch (VistaDBException e)
            {
                if (WriteExceptionsToEventLog)
                {
                    WriteToEventLog(e, "ValidateUser");
                    throw new ProviderException(exceptionMessage);
                }
                else
                    throw e;
            }
            finally
            {
                if (reader != null)
                    reader.Close();
                conn.Close();
            }

            return isValid;
        }

        private bool CheckPassword(string password, string dbpassword)
        {
            string pass1 = password.Trim();
            string pass2 = dbpassword;

            switch (PasswordFormat)
            {
                case MembershipPasswordFormat.Encrypted:
                    pass2 = UnEncodePassword(dbpassword);
                    break;
                case MembershipPasswordFormat.Hashed:
                    pass1 = EncodePassword(password);
                    break;
            }

            if (pass1 == pass2)
                return true;

            return false;
        }

        private string EncodePassword(string password)
        {
            string encodedPassword = password;

            switch (PasswordFormat)
            {
                case MembershipPasswordFormat.Clear:
                case MembershipPasswordFormat.Encrypted:
                    encodedPassword = Convert.ToBase64String(EncryptPassword(Encoding.Unicode.GetBytes(password)));
                    break;
                case MembershipPasswordFormat.Hashed:
                    HMACSHA1 hash = new HMACSHA1();
                    hash.Key = HexToByte(machineKey.ValidationKey);
                    encodedPassword = Convert.ToBase64String(hash.ComputeHash(Encoding.Unicode.GetBytes(password)));
                    break;
                default:
                    throw new ProviderException("Unsupported password format.");
            }

            return encodedPassword;
        }

        private string GetConfigValue(string configValue, string defaultValue)
        {
            if (string.IsNullOrEmpty(configValue))
                return defaultValue;

            return configValue;
        }

        private MembershipUser GetUserFromReader(VistaDBDataReader reader)
        {
            object providerUserKey = reader.GetValue(0);
            string username = reader.GetString(1);
            string email = reader.GetString(2);

            string passwordQuestion = "";
            if (reader.GetValue(3) != DBNull.Value)
                passwordQuestion = reader.GetString(3);

            string comment = "";
            if (reader.GetValue(4) != DBNull.Value)
                comment = reader.GetString(4);

            bool isApproved = reader.GetBoolean(5);
            bool isLockedOut = reader.GetBoolean(6);
            DateTime creationDate = reader.GetDateTime(7);

            DateTime lastLoginDate = new DateTime();
            if (reader.GetValue(8) != DBNull.Value)
                lastLoginDate = reader.GetDateTime(8);

            DateTime lastActivityDate = reader.GetDateTime(9);
            DateTime lastPasswordChangedDate = reader.GetDateTime(10);

            DateTime lastLockedOutDate = new DateTime();
            if (reader.GetValue(11) != DBNull.Value)
                lastLockedOutDate = reader.GetDateTime(11);

            MembershipUser u = new MembershipUser(this.Name,
                                                  username,
                                                  providerUserKey,
                                                  email,
                                                  passwordQuestion,
                                                  comment,
                                                  isApproved,
                                                  isLockedOut,
                                                  creationDate,
                                                  lastLoginDate,
                                                  lastActivityDate,
                                                  lastPasswordChangedDate,
                                                  lastLockedOutDate);

            return u;
        }

        private byte[] HexToByte(string hexString)
        {
            byte[] returnBytes = new byte[(hexString.Length / 2) - 1];
            for (int i = 0; i < returnBytes.Length - 1; i++)
            {
                returnBytes[i] = Convert.ToByte(hexString.Substring(i * 2, 2), 16);
            }
            return returnBytes;
        }

        private string UnEncodePassword(string encodedPassword)
        {
            string password = encodedPassword;

            switch (PasswordFormat)
            {
                case MembershipPasswordFormat.Clear:
                case MembershipPasswordFormat.Encrypted:
                    password = Encoding.Unicode.GetString(DecryptPassword(Convert.FromBase64String(password)));
                    break;
                case MembershipPasswordFormat.Hashed:
                    throw new ProviderException("Cannot unencode a hashed password.");
                default:
                    throw new ProviderException("Unsupported password format.");
            }

            return password;
        }

        private void UpdateFailureCount(string username, string failureType)
        {
            VistaDBConnection conn = new VistaDBConnection(connectionString);
            VistaDBCommand cmd = new VistaDBCommand("SELECT FailedPasswordAttemptCount, " +
                                              "  FailedPasswordAttemptWindowStart, " +
                                              "  FailedPasswordAnswerAttemptCount, " +
                                              "  FailedPasswordAnswerAttemptWindowStart " +
                                              "  FROM Users  " +
                                              "  WHERE Username = @Username AND ApplicationName = @ApplicationName", conn);

            cmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = username.Trim();
            cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

            VistaDBDataReader reader = null;
            DateTime windowStart = new DateTime();
            int failureCount = 0;

            try
            {
                conn.Open();
                reader = cmd.ExecuteReader(CommandBehavior.SingleRow);

                if (reader.HasRows)
                {
                    reader.Read();

                    if (failureType == "password")
                    {
                        failureCount = (int)reader["FailedPasswordAttemptCount"];
                        windowStart = (DateTime)reader["FailedPasswordAttemptWindowStart"];
                    }
                    if (failureType == "passwordAnswer")
                    {
                        failureCount = (int)reader["FailedPasswordAnswerAttemptCount"];
                        windowStart = (DateTime)reader["FailedPasswordAnswerAttemptWindowStart"];
                    }
                }

                reader.Close();

                DateTime windowEnd = windowStart.AddMinutes(PasswordAttemptWindow);

                if (failureCount == 0 || DateTime.Now > windowEnd)
                {
                    if (failureType == "password")
                        cmd.CommandText = "UPDATE Users  " +
                                          "  SET FailedPasswordAttemptCount = @Count, " +
                                          "      FailedPasswordAttemptWindowStart = @WindowStart " +
                                          "  WHERE Username = @Username AND ApplicationName = @ApplicationName";

                    if (failureType == "passwordAnswer")

                        cmd.CommandText = "UPDATE Users  " +
                                          "  SET FailedPasswordAnswerAttemptCount = @Count, " +
                                          "      FailedPasswordAnswerAttemptWindowStart = @WindowStart " +
                                          "  WHERE Username = @Username AND ApplicationName = @ApplicationName";

                    cmd.Parameters.Clear();

                    cmd.Parameters.Add("@Count", VistaDBType.Int).Value = 1;
                    cmd.Parameters.Add("@WindowStart", VistaDBType.DateTime).Value = DateTime.Now;
                    cmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = username;
                    cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

                    if (cmd.ExecuteNonQuery() < 0)
                        throw new ProviderException("Unable to update failure count and window start.");
                }
                else
                {
                    failureCount += 1;

                    if (failureCount >= MaxInvalidPasswordAttempts)
                    {
                        cmd.CommandText = "UPDATE Users  " +
                                          "  SET IsLockedOut = @IsLockedOut, LastLockedOutDate = @LastLockedOutDate " +
                                          "  WHERE Username = @Username AND ApplicationName = @ApplicationName";

                        cmd.Parameters.Clear();

                        cmd.Parameters.Add("@IsLockedOut", VistaDBType.Bit).Value = true;
                        cmd.Parameters.Add("@LastLockedOutDate", VistaDBType.DateTime).Value = DateTime.Now;
                        cmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = username;
                        cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

                        if (cmd.ExecuteNonQuery() < 0)
                            throw new ProviderException("Unable to lock out user.");
                    }
                    else
                    {
                       if (failureType == "password")
                            cmd.CommandText = "UPDATE Users  " +
                                              "  SET FailedPasswordAttemptCount = @Count" +
                                              "  WHERE Username = @Username AND ApplicationName = @ApplicationName";

                        if (failureType == "passwordAnswer")
                            cmd.CommandText = "UPDATE Users  " +
                                              "  SET FailedPasswordAnswerAttemptCount = @Count" +
                                              "  WHERE Username = @Username AND ApplicationName = @ApplicationName";

                        cmd.Parameters.Clear();

                        cmd.Parameters.Add("@Count", VistaDBType.Int).Value = failureCount;
                        cmd.Parameters.Add("@Username", VistaDBType.VarChar, 255).Value = username;
                        cmd.Parameters.Add("@ApplicationName", VistaDBType.VarChar, 255).Value = pApplicationName;

                        if (cmd.ExecuteNonQuery() < 0)
                            throw new ProviderException("Unable to update failure count.");
                    }
                }
            }
            catch (VistaDBException e)
            {
                if (WriteExceptionsToEventLog)
                {
                    WriteToEventLog(e, "UpdateFailureCount");

                    throw new ProviderException(exceptionMessage);
                }
                else
                    throw e;
            }
            finally
            {
                if (reader != null)
                    reader.Close();
                conn.Close();
            }
        }

        private void WriteToEventLog(VistaDBException e, string action)
        {
            EventLog log = new EventLog();
            log.Source = eventSource;
            log.Log = eventLog;

            string message = "An exception occurred communicating with the data source." + "\n";
            message += "Action: " + action + "\n";
            message += "Exception: " + e.ToString();

            try
            {
                log.WriteEntry(message);
            }
            catch (Exception ex)
            {
                throw;
            }
        }
    }
}